import pytest

from contextlib import contextmanager
import math
from pathlib import Path

# import fpdf
from fpdf.drawing import (
    Point,
    Transform,
    IntersectionRule,
    GraphicsStyle,
    Move,
    Line,
    BezierCurve,
    QuadraticBezierCurve,
    Arc as A,
    ImplicitClose,
    Close,
    RoundedRectangle,
    Ellipse,
)

from fpdf.svg import (
    resolve_length,
)

SVG_SOURCE_DIR = Path(__file__).resolve().parent / "svg_sources"


@contextmanager
def no_error():
    yield


def svgfile(*names):
    return SVG_SOURCE_DIR.joinpath(*names)


def P(x, y):
    return Point(float(x), float(y))


def pointifier(source_fn):
    def wrapper(*args):
        return source_fn(*(P(arg1, arg2) for arg1, arg2 in zip(args[::2], args[1::2])))

    return wrapper


M = pointifier(Move)
L = pointifier(Line)
C = pointifier(BezierCurve)
Q = pointifier(QuadraticBezierCurve)
iz = pointifier(ImplicitClose)
Z = pointifier(Close)

Re = pointifier(RoundedRectangle)
El = pointifier(Ellipse)


def Gs(**kwargs):
    style = GraphicsStyle()
    # this is set on all stylables by the SVG machinery
    style.auto_close = False
    for name, val in kwargs.items():
        setattr(style, name, val)

    return style


test_svg_shape_tags = (
    pytest.param(
        '<rect x="20" y="20" width="25.4mm" height="2.54 cm"/>',
        [M(0, 0), Re(20, 20, 72, 72, 0, 0)],
        no_error(),
        id="rect",
    ),
    pytest.param(
        '<rect x="20" y="20" width="60" height="60"/>',
        [M(0, 0), Re(20, 20, 60, 60, 0, 0)],
        no_error(),
        id="rect",
    ),
    pytest.param(
        '<rect x="20" y="20" width="60" height="60" rx="none" ry="none"/>',
        [M(0, 0), Re(20, 20, 60, 60, 0, 0)],
        no_error(),
        id="rect rx and ry none",
    ),
    pytest.param(
        '<rect x="20" y="20" width="60" height="60" rx="10" ry="none"/>',
        [M(0, 0), Re(20, 20, 60, 60, 10, 0)],
        no_error(),
        id="rect rx ry none",
    ),
    pytest.param(
        '<rect x="20" y="20" width="60" height="60" rx="10"/>',
        [M(0, 0), Re(20, 20, 60, 60, 10, 10)],
        no_error(),
        id="rect rx no ry",
    ),
    pytest.param(
        '<rect x="20" y="20" width="60" height="60" rx="10" ry="auto"/>',
        [M(0, 0), Re(20, 20, 60, 60, 10, 10)],
        no_error(),
        id="rect rx ry auto",
    ),
    pytest.param(
        '<rect x="20" y="20" width="60" height="60" ry="30"/>',
        [M(0, 0), Re(20, 20, 60, 60, 30, 30)],
        no_error(),
        id="rect ry no rx",
    ),
    pytest.param(
        '<rect x="20" y="20" width="60" height="60" rx="none" ry="30"/>',
        [M(0, 0), Re(20, 20, 60, 60, 0, 30)],
        no_error(),
        id="rect ry rx none",
    ),
    pytest.param(
        '<rect x="20" y="20" width="60" height="60" rx="auto" ry="30"/>',
        [M(0, 0), Re(20, 20, 60, 60, 30, 30)],
        no_error(),
        id="rect ry rx auto",
    ),
    pytest.param(
        '<rect x="20" y="20" width="500" height="60" rx="100" ry="10"/>',
        [M(0, 0), Re(20, 20, 500, 60, 100, 10)],
        no_error(),
        id="rect rx and ry",
    ),
    pytest.param(
        '<rect x="20" y="20" width="0" height="60" rx="10" ry="10"/>',
        [],
        no_error(),
        id="rect width 0",
    ),
    pytest.param(
        '<rect x="20" y="20" width="60" height="60" rx="60" ry="60"/>',
        [M(0, 0), Re(20, 20, 60, 60, 30, 30)],
        no_error(),
        id="rect overlarge rx and ry",
    ),
    pytest.param(  # spotted in issue #720
        '<rect x="1mm" y="1mm" width="100%" height="100%"/>',
        [M(0.0, 0.0), Re(2.834645669291339, 2.834645669291339, 100, 100, 0, 0)],
        no_error(),
        id="rect with mm and percents",
    ),
    pytest.param(
        '<rect x="20" y="20" width="500" height="60" rx="-100" ry="10"/>',
        [],
        pytest.raises(ValueError),
        id="rect negative rx",
    ),
    pytest.param(
        '<rect x="20" y="20" width="-500" height="60" rx="100" ry="10"/>',
        [],
        pytest.raises(ValueError),
        id="rect negative width",
    ),
    pytest.param(
        '<circle r="10"/>',
        [M(0, 0), El(10, 10, 0, 0)],
        no_error(),
        id="circle no cx no cy",
    ),
    pytest.param(
        '<circle cx="10" r="10"/>',
        [M(0, 0), El(10, 10, 10, 0)],
        no_error(),
        id="circle cx no cy",
    ),
    pytest.param(
        '<circle cy="10" r="10"/>',
        [M(0, 0), El(10, 10, 0, 10)],
        no_error(),
        id="circle cy no cx",
    ),
    pytest.param(
        '<circle cx="10" cy="20" r="10"/>',
        [M(0, 0), El(10, 10, 10, 20)],
        no_error(),
        id="circle cy cx",
    ),
    pytest.param(
        '<circle r="-10"/>',
        [M(0, 0), El(-10, -10, 0, 0)],
        no_error(),
        id="circle negative r",
    ),
    pytest.param(
        '<circle cx="10" cy="10"/>', [], pytest.raises(KeyError), id="circle no r"
    ),
    pytest.param(
        '<ellipse rx="10"/>',
        [M(0, 0), El(10, 10, 0, 0)],
        no_error(),
        id="ellipse no cx no cy no ry",
    ),
    pytest.param(
        '<ellipse rx="10" ry="auto"/>',
        [M(0, 0), El(10, 10, 0, 0)],
        no_error(),
        id="ellipse no cx no cy ry auto",
    ),
    pytest.param(
        '<ellipse ry="10"/>',
        [M(0, 0), El(10, 10, 0, 0)],
        no_error(),
        id="ellipse no cx no cy no rx",
    ),
    pytest.param(
        '<ellipse rx="auto" ry="10"/>',
        [M(0, 0), El(10, 10, 0, 0)],
        no_error(),
        id="ellipse no cx no cy rx auto",
    ),
    pytest.param(
        "<ellipse/>",
        [],
        no_error(),
        id="ellipse empty",
    ),
    pytest.param(
        '<ellipse cx="10" rx="10"/>',
        [M(0, 0), El(10, 10, 10, 0)],
        no_error(),
        id="ellipse cx no cy",
    ),
    pytest.param(
        '<ellipse cy="10" rx="10"/>',
        [M(0, 0), El(10, 10, 0, 10)],
        no_error(),
        id="ellipse cy no cx",
    ),
    pytest.param(
        '<ellipse cx="10" cy="20" rx="10"/>',
        [M(0, 0), El(10, 10, 10, 20)],
        no_error(),
        id="ellipse cy cx",
    ),
    pytest.param(
        '<ellipse rx="-10"/>',
        [M(0, 0), El(-10, -10, 0, 0)],
        no_error(),
        id="ellipse negative r",
    ),
    pytest.param(
        '<ellipse rx="-10"/>',
        [M(0, 0), El(-10, -10, 0, 0)],
        no_error(),
        id="ellipse negative r",
    ),
    pytest.param(
        '<line x1="0" y1="0" x2="10" y2="10"/>',
        [M(0, 0), L(10, 10)],
        no_error(),
        id="line",
    ),
    pytest.param(
        '<line y1="0" x2="10" y2="10"/>',
        [],
        pytest.raises(KeyError),
        id="line no x1",
    ),
    pytest.param(
        '<polyline points="1, 0 10, 10, -20, -50"/>',
        [M(1, 0), L(10, 10), L(-20, -50)],
        no_error(),
        id="polyline",
    ),
    pytest.param(
        "<polyline/>",
        [],
        pytest.raises(KeyError),
        id="polyline no points",
    ),
    pytest.param(
        '<polygon points="1, 0 10, 10, -20, -50"/>',
        [M(1, 0), L(10, 10), L(-20, -50), Z()],
        no_error(),
        id="polygon",
    ),
    pytest.param(
        "<polygon/>",
        [],
        pytest.raises(KeyError),
        id="polygon no points",
    ),
    pytest.param(
        '<circle r="10" fill="rgb(100, 200)"/>',
        [],
        pytest.raises(ValueError),
        id="invalid rgb() color",
    ),
)

test_svg_transforms = (
    pytest.param(
        "matrix(1,2,3,4,5,6)",
        Transform(1, 2, 3, 4, 5, 6),
        no_error(),
        id="matrix",
    ),
    pytest.param(
        "rotate(30)",
        Transform.rotation_d(30),
        no_error(),
        id="rotate",
    ),
    pytest.param(
        "rotate(30, 10, 10)",
        Transform.rotation_d(30).about(10, 10),
        no_error(),
        id="rotate about",
    ),
    pytest.param(
        "rotate(30, 10)",
        Transform.identity(),
        pytest.raises(ValueError),
        id="rotate bad syntax",
    ),
    pytest.param(
        "scale(2)",
        Transform.scaling(x=2, y=2),
        no_error(),
        id="scale combined",
    ),
    pytest.param(
        "scale(2, 1)",
        Transform.scaling(x=2, y=1),
        no_error(),
        id="scale x",
    ),
    pytest.param(
        "scale(1 2)",
        Transform.scaling(x=1, y=2),
        no_error(),
        id="scale y",
    ),
    pytest.param(
        "scale(1 2 3)",
        Transform.identity(),
        pytest.raises(ValueError),
        id="scale bad syntax",
    ),
    pytest.param(
        "scaleX(2)",
        Transform.scaling(x=2, y=1),
        no_error(),
        id="scaleX",
    ),
    pytest.param(
        "scaleY(2)",
        Transform.scaling(x=1, y=2),
        no_error(),
        id="scaleY",
    ),
    pytest.param(
        "skew(2)",
        Transform.shearing(x=math.tan(math.radians(2)), y=0),
        no_error(),
        id="skew x-only",
    ),
    pytest.param(
        "skew(3, 3)",
        Transform.shearing(x=math.tan(math.radians(3))),
        no_error(),
        id="skew x and y",
    ),
    pytest.param(
        "skew(2, 3, 4, 5, 6)",
        None,
        pytest.raises(ValueError),
        id="skew too many args",
    ),
    pytest.param(
        "skewX(2)",
        Transform.shearing(x=math.tan(math.radians(2)), y=0),
        no_error(),
        id="skewX",
    ),
    pytest.param(
        "skewY(2)",
        Transform.shearing(x=0, y=math.tan(math.radians(2))),
        no_error(),
        id="skewY",
    ),
    pytest.param(
        "translate(20)",
        Transform.translation(x=20, y=0),
        no_error(),
        id="translate x-only",
    ),
    pytest.param(
        "translate(20, 30)",
        Transform.translation(x=20, y=30),
        no_error(),
        id="translate x and y",
    ),
    pytest.param(
        "translate(20, 30, 45)",
        None,
        pytest.raises(ValueError),
        id="translate too many args",
    ),
    pytest.param(
        "translateX(10)",
        Transform.translation(x=10, y=0),
        no_error(),
        id="translateX",
    ),
    pytest.param(
        "translateY(10)",
        Transform.translation(x=0, y=10),
        no_error(),
        id="translateY",
    ),
    pytest.param(
        "skewX(30) scale(1, 1.25) translate(200, 200) rotate(45) translate(-500, -500)",
        Transform.translation(-500, -500)
        .rotate_d(45)
        .translate(200, 200)
        .scale(1, 1.25)
        .shear(math.tan(math.radians(30)), 0),
        no_error(),
        id="multiple",
    ),
)

test_svg_transform_documents = (
    pytest.param(svgfile("transforms", "matrix.svg"), id="matrix"),
    pytest.param(svgfile("transforms", "rotate.svg"), id="rotate"),
    pytest.param(svgfile("transforms", "scale.svg"), id="scale"),
    pytest.param(svgfile("transforms", "skew.svg"), id="skew"),
    pytest.param(svgfile("transforms", "translate.svg"), id="translate"),
    pytest.param(svgfile("transforms", "multi.svg"), id="multiple"),
)

test_svg_attribute_conversion = (
    pytest.param(
        '<path fill="none"/>',
        Gs(fill_color=None),
        no_error(),
        id="fill color none",
    ),
    pytest.param(
        '<path fill="inherit"/>',
        Gs(fill_color=GraphicsStyle.INHERIT),
        no_error(),
        id="fill color inherit",
    ),
    pytest.param(
        '<path fill="black"/>',
        Gs(fill_color="#000"),
        no_error(),
        id="fill color name",
    ),
    pytest.param(
        '<path fill="#0007"/>',
        Gs(fill_color="#0007", fill_opacity=0x77 / 0xFF),
        no_error(),
        id="fill color rgba",
    ),
    pytest.param(
        '<path fill="1"/>',
        Gs(),
        pytest.raises(ValueError),
        id="fill color invalid",
    ),
    pytest.param(
        '<path fill-rule="inherit"/>',
        Gs(intersection_rule=GraphicsStyle.INHERIT),
        no_error(),
        id="fill-rule inherit",
    ),
    pytest.param(
        '<path fill-rule="nonzero"/>',
        Gs(intersection_rule=IntersectionRule.NONZERO),
        no_error(),
        id="fill-rule nonzero",
    ),
    pytest.param(
        '<path fill-rule="evenodd"/>',
        Gs(intersection_rule=IntersectionRule.EVENODD),
        no_error(),
        id="fill-rule evenodd",
    ),
    pytest.param(
        '<path fill-rule="none"/>',
        Gs(),
        pytest.raises(ValueError),
        id="fill-rule invalid",
    ),
    pytest.param(
        '<path fill-opacity="0.5"/>',
        Gs(fill_opacity=0.5),
        no_error(),
        id="fill-opacity 0.5",
    ),
    pytest.param(
        '<path fill-opacity="-2"/>',
        Gs(fill_opacity=0.0),
        no_error(),
        id="fill-opacity too small",
    ),
    pytest.param(
        '<path fill-opacity="5"/>',
        Gs(fill_opacity=1.0),
        no_error(),
        id="fill-opacity too big",
    ),
    pytest.param(
        '<path fill-opacity="inherit"/>',
        Gs(fill_opacity=GraphicsStyle.INHERIT),
        no_error(),
        id="fill-opacity inherit",
    ),
    pytest.param(
        '<path fill-opacity="none"/>',
        Gs(),
        pytest.raises(ValueError),
        id="fill-opacity invalid",
    ),
    pytest.param(
        '<path stroke="none"/>',
        Gs(stroke_color=None),
        no_error(),
        id="stroke color none",
    ),
    pytest.param(
        '<path stroke="inherit"/>',
        Gs(stroke_color=GraphicsStyle.INHERIT),
        no_error(),
        id="stroke color inherit",
    ),
    pytest.param(
        '<path stroke="black"/>',
        Gs(stroke_color="#000"),
        no_error(),
        id="stroke color name",
    ),
    pytest.param(
        '<path stroke="#0007"/>',
        Gs(stroke_color="#0007", stroke_opacity=0x77 / 0xFF),
        no_error(),
        id="stroke color rgba",
    ),
    pytest.param(
        '<path stroke="1"/>',
        Gs(),
        pytest.raises(ValueError),
        id="stroke color invalid",
    ),
    pytest.param(
        '<path stroke-width="0"/>',
        Gs(stroke_width=None),
        no_error(),
        id="stroke-width 0",
    ),
    pytest.param(
        '<path stroke-width="2"/>',
        Gs(stroke_width=2),
        no_error(),
        id="stroke-width number",
    ),
    pytest.param(  # issue #526
        '<path stroke-width="2px"/>',
        Gs(stroke_width=2 * 0.75),
        no_error(),
        id="stroke-width number",
    ),
    pytest.param(
        '<path stroke-width="inherit"/>',
        Gs(),
        no_error(),
        id="stroke-width inherit",
    ),
    pytest.param(
        '<path stroke-width="-2"/>',
        Gs(),
        pytest.raises(ValueError),
        id="stroke-width negative",
    ),
    pytest.param(
        '<path stroke-width="bad"/>',
        Gs(),
        pytest.raises(ValueError),
        id="stroke-width invalid",
    ),
    pytest.param(
        '<path stroke-dasharray="1 2 3 4 5"/>',
        Gs(stroke_dash_pattern=[1, 2, 3, 4, 5]),
        no_error(),
        id="stroke-dasharray list",
    ),
    pytest.param(
        '<path stroke-dasharray="1"/>',
        Gs(stroke_dash_pattern=[1]),
        no_error(),
        id="stroke-dasharray single value",
    ),
    pytest.param(
        '<path stroke-dasharray="inherit"/>',
        Gs(),
        no_error(),
        id="stroke-dasharray inherit",
    ),
    pytest.param(
        '<path stroke-dasharray="bad"/>',
        Gs(),
        pytest.raises(ValueError),
        id="stroke-dasharray invalid",
    ),
    pytest.param(
        '<path stroke-dasharray="1" stroke-dashoffset="1"/>',
        Gs(stroke_dash_pattern=[1], stroke_dash_phase=1),
        no_error(),
        id="stroke-dashoffset",
    ),
    pytest.param(
        '<path stroke-linecap="butt"/>',
        Gs(stroke_cap_style="butt"),
        no_error(),
        id="stroke-linecap butt",
    ),
    pytest.param(
        '<path stroke-linecap="round"/>',
        Gs(stroke_cap_style="round"),
        no_error(),
        id="stroke-linecap round",
    ),
    pytest.param(
        '<path stroke-linecap="square"/>',
        Gs(stroke_cap_style="square"),
        no_error(),
        id="stroke-linecap square",
    ),
    pytest.param(
        '<path stroke-linecap="inherit"/>',
        Gs(stroke_cap_style=GraphicsStyle.INHERIT),
        no_error(),
        id="stroke-linecap inherit",
    ),
    pytest.param(
        '<path stroke-linecap="bad"/>',
        Gs(),
        pytest.raises(ValueError),
        id="stroke-linecap invalid",
    ),
    pytest.param(
        '<path stroke-linejoin="miter"/>',
        Gs(stroke_join_style="miter"),
        no_error(),
        id="stroke-linejoin miter",
    ),
    pytest.param(
        '<path stroke-linejoin="round"/>',
        Gs(stroke_join_style="round"),
        no_error(),
        id="stroke-linejoin round",
    ),
    pytest.param(
        '<path stroke-linejoin="bevel"/>',
        Gs(stroke_join_style="bevel"),
        no_error(),
        id="stroke-linejoin bevel",
    ),
    pytest.param(
        '<path stroke-linejoin="inherit"/>',
        Gs(stroke_join_style=GraphicsStyle.INHERIT),
        no_error(),
        id="stroke-linejoin inherit",
    ),
    pytest.param(
        '<path stroke-linejoin="bad"/>',
        Gs(),
        pytest.raises(ValueError),
        id="stroke-linejoin invalid",
    ),
    pytest.param(
        '<path stroke-miterlimit="2"/>',
        Gs(stroke_miter_limit=2),
        no_error(),
        id="stroke-miterlimit",
    ),
    pytest.param(
        '<path stroke-miterlimit="inherit"/>',
        Gs(stroke_miter_limit=GraphicsStyle.INHERIT),
        no_error(),
        id="stroke-miterlimit inherit",
    ),
    pytest.param(
        '<path stroke-miterlimit="0.5"/>',
        Gs(),
        pytest.raises(ValueError),
        id="stroke-miterlimit too small",
    ),
    pytest.param(
        '<path stroke-miterlimit="bad"/>',
        Gs(),
        pytest.raises(ValueError),
        id="stroke-miterlimit invalid",
    ),
    pytest.param(
        '<path stroke-opacity="0.5"/>',
        Gs(stroke_opacity=0.5),
        no_error(),
        id="stroke-opacity 0.5",
    ),
    pytest.param(
        '<path stroke-opacity="-2"/>',
        Gs(stroke_opacity=0.0),
        no_error(),
        id="stroke-opacity too small",
    ),
    pytest.param(
        '<path stroke-opacity="5"/>',
        Gs(stroke_opacity=1.0),
        no_error(),
        id="stroke-opacity too big",
    ),
    pytest.param(
        '<path stroke-opacity="inherit"/>',
        Gs(stroke_opacity=GraphicsStyle.INHERIT),
        no_error(),
        id="stroke-opacity inherit",
    ),
    pytest.param(
        '<path stroke-opacity="none"/>',
        Gs(),
        pytest.raises(ValueError),
        id="stroke-opacity invalid",
    ),
    pytest.param(
        '<path stroke-dasharray="" opacity="" transform=""/>',
        Gs(),
        no_error(),
        id="empty attribute values",
    ),
)

test_svg_sources = (
    pytest.param(svgfile("arcs01.svg"), id="SVG spec arcs01"),
    pytest.param(svgfile("arcs02.svg"), id="SVG spec arcs02"),
    pytest.param(svgfile("circle01.svg"), id="SVG spec circle01"),
    pytest.param(svgfile("cubic01.svg"), id="SVG spec cubic01 (modified)"),
    pytest.param(svgfile("cubic02.svg"), id="SVG spec cubic02 (modified)"),
    pytest.param(svgfile("ellipse01.svg"), id="SVG spec ellipse01"),
    pytest.param(svgfile("line01.svg"), id="SVG spec line01"),
    pytest.param(svgfile("polygon01.svg"), id="SVG spec polygon01"),
    pytest.param(svgfile("polyline01.svg"), id="SVG spec polyline01"),
    pytest.param(svgfile("quad01.svg"), id="SVG spec quad01"),
    pytest.param(svgfile("rect01.svg"), id="SVG spec rect01"),
    pytest.param(svgfile("rect02.svg"), id="SVG spec rect02"),
    pytest.param(svgfile("triangle01.svg"), id="SVG spec triangle01"),
    pytest.param(svgfile("SVG_logo.svg"), id="SVG logo from wikipedia"),
    pytest.param(svgfile("viewbox.svg"), id="weird viewbox"),
    pytest.param(svgfile("search.svg"), id="search icon"),  # issue 356
    # discovered while investigating issue 358:
    pytest.param(svgfile("issue_358b.svg"), id="repeated relative move"),
    pytest.param(svgfile("issue_358.svg"), id="arc start & initial point"),  # issue 358
    pytest.param(svgfile("Ghostscript_colorcircle.svg"), id="ghostscript colorcircle"),
    pytest.param(svgfile("Ghostscript_escher.svg"), id="ghostscript escher"),
    pytest.param(svgfile("use-xlink-href.svg"), id="use xlink:href - issue #446"),
    pytest.param(svgfile("rgb-color-issue-480.svg"), id="rgb() color - issue #480"),
    pytest.param(
        svgfile("shapes_def_test.svg"), id="shapes defined in 'defs' tag - issue #858"
    ),
    pytest.param(svgfile("clip_path.svg"), id="clip path - issue #858"),
    pytest.param(svgfile("embedded-raster-images.svg"), id="Raster <image> tests"),
    pytest.param(
        svgfile("use-image-def.svg"),
        id="Use xlink:href to insert an <image> from <defs>",
    ),
    pytest.param(
        svgfile("issue-1076.svg"),
        id="<clipPath> containing a <path>",
    ),
    pytest.param(
        svgfile("path_clippingpath.svg"),
        id="<clipPath> containing a <path> used in a group with color - issue #1147",
    ),
    pytest.param(svgfile("text-samples.svg"), id="<text> tests"),
    pytest.param(svgfile("text01.svg"), id="<text> test"),
    pytest.param(svgfile("tspan01.svg"), id="<text> with <tspan>"),
    pytest.param(svgfile("tspan02.svg"), id="<text> with multiple <tspan>"),
    pytest.param(svgfile("gradient_linear.svg"), id="Linear gradient"),
    pytest.param(svgfile("gradient_radial.svg"), id="Radial gradient"),
    pytest.param(svgfile("gradient_multiple.svg"), id="Multiple gradients"),
    pytest.param(svgfile("gradient_opacity.svg"), id="Gradient with opacity"),
    pytest.param(svgfile("gradient_spread_methods.svg"), id="Gradient spread methods"),
)

svg_path_edge_cases = (
    pytest.param(
        "M0 1L2 3z", [M(0, 1), L(2, 3), Z()], id="no whitespace around commands"
    ),
    pytest.param(
        " M    0   1  L  2   3 z  ", [M(0, 1), L(2, 3), Z()], id="extra whitespace"
    ),
    pytest.param("M0,1l2,3", [M(0, 1), L(2, 4)], id="comma separation"),
    pytest.param("M 0 , 1 L 2 , 3", [M(0, 1), L(2, 3)], id="commas and spaces"),
    pytest.param("M 0,1 L-2-3", [M(0, 1), L(-2, -3)], id="negative number separation"),
    pytest.param("M 0,1 L+2+3", [M(0, 1), L(2, 3)], id="unary plus number separation"),
    pytest.param("M 0 1 2 3 4 5", [M(0, 1), L(2, 3), L(4, 5)], id="implicit L"),
    pytest.param("m 0 1 2 3 4 5", [M(0, 1), L(2, 4), L(6, 9)], id="implicit l"),
    pytest.param(
        "m 0. .1 L 2.2 3.3",
        [M(0, 0.1), L(2.2, 3.3)],
        id="floating point numbers",
    ),
    pytest.param("M0..1L.2.3.4.5", [M(0.0, 0.1), L(0.2, 0.3), L(0.4, 0.5)], id="why"),
)

svg_path_directives = (
    pytest.param("M 0 1 L 2 3", [M(0, 1), L(2, 3)], id="line"),
    pytest.param("m 0 1 l 2 3", [M(0, 1), L(2, 4)], id="relative line"),
    pytest.param("M 0 1 H 2", [M(0, 1), L(2, 1)], id="horizontal line"),
    pytest.param("M 0 1 h 2", [M(0, 1), L(2, 1)], id="relative horizontal line"),
    pytest.param("M 0 1 V 2", [M(0, 1), L(0, 2)], id="vertical line"),
    pytest.param("M 0 1 v 2", [M(0, 1), L(0, 3)], id="relative vertical line"),
    pytest.param(
        "M 0 1 C 2 3 4 5 6 7", [M(0, 1), C(2, 3, 4, 5, 6, 7)], id="cubic bezier"
    ),
    pytest.param(
        "M 0 1 c 2 3 4 5 6 7",
        [M(0, 1), C(2, 4, 4, 6, 6, 8)],
        id="relative cubic bezier",
    ),
    pytest.param("M 0 1 Q 2 3 4 5", [M(0, 1), Q(2, 3, 4, 5)], id="quadratic bezier"),
    pytest.param(
        "M 0 1 q 2 3 4 5", [M(0, 1), Q(2, 4, 4, 6)], id="relative quadratic bezier"
    ),
    pytest.param(
        "M 0 1 A 2 3 0 1 0 4 5",
        [M(0, 1), A(P(2, 3), 0, True, False, P(4, 5))],
        id="arc",
    ),
    pytest.param(
        "M 0 1 a 2 3 0 1 0 4 5",
        [M(0, 1), A(P(2, 3), 0, True, False, P(4, 6))],
        id="relative arc",
    ),
    pytest.param(
        "M 0 1 C 2 3 4 5 6 7 S 8 9 10 11",
        [M(0, 1), C(2, 3, 4, 5, 6, 7), C(8, 9, 8, 9, 10, 11)],
        id="smooth cubic bezier",
    ),
    pytest.param(
        "M 0 1 C 2 3 4 5 6 7 s 8 9 10 11",
        [M(0, 1), C(2, 3, 4, 5, 6, 7), C(8, 9, 14, 16, 16, 18)],
        id="relative smooth cubic bezier",
    ),
    pytest.param(
        "M 0 1 Q 2 3 4 5 T 6 7",
        [M(0, 1), Q(2, 3, 4, 5), Q(6, 7, 6, 7)],
        id="smooth quadratic bezier",
    ),
    pytest.param(
        "M 0 1 Q 2 3 4 5 t 6 7",
        [M(0, 1), Q(2, 3, 4, 5), Q(6, 7, 10, 12)],
        id="relative smooth quadratic bezier",
    ),
    pytest.param("M 0 1 z", [M(0, 1), Z()], id="close"),
    pytest.param(
        "M 0 1 L 2 3 M 3 4 L 5 6",
        [M(0, 1), L(2, 3), iz(), M(3, 4), L(5, 6)],
        id="implicit close",
    ),
    pytest.param(
        "M 0 0 c -7e-5 -4e-5 -8.8492 -3.1382 -8.8493 -3.1383",
        [M(0, 0), C(-7e-5, -4e-5, -8.8492, -3.1382, -8.8493, -3.1383)],
        id="exponentiated numbers",  # cf. issue #376
    ),
    pytest.param(
        "M 0 1 a5.168 5.168 0 013.06-1.837",
        [M(0, 1), A(P(5.168, 5.168), 0.0, False, True, P(3.06, -0.837))],
        id="arc with joined arguments",  # cf. issue #450
    ),
)

svg_path_render_tests = (
    pytest.param(
        "M 0 1 C 2 3 4 5 6 7 S 8 9 10 11",
        "q 0 1 m 2 3 4 5 6 7 c 8 9 8 9 10 11 c h B Q",
        id="smooth cubic bezier",
    ),
    pytest.param(
        "M 0 1 S 8 9 10 11",
        "q 0 1 m 0 1 8 9 10 11 c h B Q",
        id="smooth cubic bezier unchained",
    ),
    pytest.param(
        "M 0 1 C 2 3 4 5 6 7 s 8 9 10 11",
        "q 0 1 m 2 3 4 5 6 7 c 8 9 14 16 16 18 c h B Q",
        id="relative smooth cubic bezier",
    ),
    pytest.param(
        "M 0 1 s 8 9 10 11",
        "q 0 1 m 0 1 8 10 10 12 c h B Q",
        id="relative smooth cubic bezier unchained",
    ),
    pytest.param(
        "M 0 1 Q 2 3 4 5 T 6 7",
        "q 0 1 m 1.3333 2.3333 2.6667 3.6667 4 5 c 5.3333 6.3333 6 7 6 7 c h B Q",
        id="smooth quadratic bezier",
    ),
    pytest.param(
        "M 0 1 T 6 7",
        "q 0 1 m 0 1 2 3 6 7 c h B Q",
        id="smooth quadratic bezier unchained",
    ),
    pytest.param(
        "M 0 1 Q 2 3 4 5 t 6 7",
        "q 0 1 m 1.3333 2.3333 2.6667 3.6667 4 5 c 5.3333 6.3333 7.3333 8.6667 10 12 c h B Q",
        id="relative smooth quadratic bezier",
    ),
    pytest.param(
        "M 0 1 t 6 7",
        "q 0 1 m 0 1 2 3.3333 6 8 c h B Q",
        id="relative smooth quadratic bezier unchained",
    ),
)

svg_path_implicit_directives = (
    pytest.param("M 0 1 L 2 3 4 5", [M(0, 1), L(2, 3), L(4, 5)], id="line"),
    pytest.param("m 0 1 l 2 3 4 5", [M(0, 1), L(2, 4), L(6, 9)], id="relative line"),
    pytest.param("M 0 1 H 2 3", [M(0, 1), L(2, 1), L(3, 1)], id="horizontal line"),
    pytest.param(
        "M 0 1 h 2 3", [M(0, 1), L(2, 1), L(5, 1)], id="relative horizontal line"
    ),
    pytest.param("M 0 1 V 2 3", [M(0, 1), L(0, 2), L(0, 3)], id="vertical line"),
    pytest.param(
        "M 0 1 v 2 3", [M(0, 1), L(0, 3), L(0, 6)], id="relative vertical line"
    ),
    pytest.param(
        "M 0 1 C 2 3 4 5 6 7 8 9 10 11 12 13",
        [M(0, 1), C(2, 3, 4, 5, 6, 7), C(8, 9, 10, 11, 12, 13)],
        id="cubic bezier",
    ),
    pytest.param(
        "M 0 1 c 2 3 4 5 6 7 8 9 10 11 12 13",
        [M(0, 1), C(2, 4, 4, 6, 6, 8), C(14, 17, 16, 19, 18, 21)],
        id="relative cubic bezier",
    ),
    pytest.param(
        "M 0 1 Q 2 3 4 5 6 7 8 9",
        [M(0, 1), Q(2, 3, 4, 5), Q(6, 7, 8, 9)],
        id="quadratic bezier",
    ),
    pytest.param(
        "M 0 1 q 2 3 4 5 6 7 8 9",
        [M(0, 1), Q(2, 4, 4, 6), Q(10, 13, 12, 15)],
        id="relative quadratic bezier",
    ),
    pytest.param(
        "M 0 1 A 2 3 0 1 0 4 5 6 7 0 1 0 8 9",
        [
            M(0, 1),
            A(P(2, 3), 0, True, False, P(4, 5)),
            A(P(6, 7), 0, True, False, P(8, 9)),
        ],
        id="arc",
    ),
    pytest.param(
        "M 0 1 a 2 3 0 1 0 4 5 6 7 0 1 0 8 9",
        [
            M(0, 1),
            A(P(2, 3), 0, True, False, P(4, 6)),
            A(P(6, 7), 0, True, False, P(12, 15)),
        ],
        id="relative arc",
    ),
    pytest.param(
        "M 0 1 C 2 3 4 5 6 7 S 8 9 10 11 12 13 14 15",
        [
            M(0, 1),
            C(2, 3, 4, 5, 6, 7),
            C(8, 9, 8, 9, 10, 11),
            C(12, 13, 12, 13, 14, 15),
        ],
        id="smooth cubic bezier",
    ),
    pytest.param(
        "M 0 1 C 2 3 4 5 6 7 s 8 9 10 11 12 13 14 15",
        [
            M(0, 1),
            C(2, 3, 4, 5, 6, 7),
            C(8, 9, 14, 16, 16, 18),
            C(18, 20, 28, 31, 30, 33),
        ],
        id="relative smooth cubic bezier",
    ),
    pytest.param(
        "M 0 1 Q 2 3 4 5 T 6 7 8 9",
        [M(0, 1), Q(2, 3, 4, 5), Q(6, 7, 6, 7), Q(6, 7, 8, 9)],
        id="smooth quadratic bezier",
    ),
    pytest.param(
        "M 0 1 Q 2 3 4 5 t 6 7 8 9",
        [M(0, 1), Q(2, 3, 4, 5), Q(6, 7, 10, 12), Q(14, 17, 18, 21)],
        id="relative smooth quadratic bezier",
    ),
)


def svg_snippet(width="", height="", viewbox="", aspect=True):
    if width:
        width = f'width="{width}"'
    if height:
        height = f'height="{height}"'
    if viewbox:
        viewbox = f'viewBox="{viewbox}"'
    if aspect:
        aspect = 'preserveAspectRatio="xMidYMid"'
    else:
        aspect = 'preserveAspectRatio="none"'

    return f"""<?xml version="1.0" standalone="no"?>
        <svg {width} {height} {viewbox} {aspect} xmlns="http://www.w3.org/2000/svg" version="1.1">
        </svg>
        """


# test "renders" these onto a 10pt by 10pt page.
svg_shape_info_tests = (
    pytest.param(
        svg_snippet(width="10mm"),
        (resolve_length("10mm"), 10),
        Transform.identity(),
        no_error(),
        id="width only",
    ),
    pytest.param(
        svg_snippet(height="10mm"),
        (10, resolve_length("10mm")),
        Transform.identity(),
        no_error(),
        id="height only",
    ),
    pytest.param(
        svg_snippet(width="100%", height="100%"),
        (10, 10),
        Transform.identity(),
        no_error(),
        id="fit percent",
    ),
    pytest.param(
        svg_snippet(width="50%", height="50%"),
        (5, 5),
        Transform.identity(),
        no_error(),
        id="small percent",
    ),
    pytest.param(
        svg_snippet(width="10mm", height="10mm"),
        (resolve_length("10mm"),) * 2,
        Transform.identity(),
        no_error(),
        id="same shape",
    ),
    pytest.param(
        svg_snippet(width="11mm", height="9mm"),
        (resolve_length("11mm"), resolve_length("9mm")),
        Transform.identity(),
        no_error(),
        id="different shape",
    ),
    pytest.param(
        svg_snippet(width="110%", height="90%"),
        (11, 9),
        Transform.identity(),
        no_error(),
        id="different shape, percent",
    ),
    pytest.param(
        svg_snippet(viewbox="0 0 100 100"),
        (10, 10),
        Transform.scaling(0.1),
        no_error(),
        id="viewbox 100",
    ),
    pytest.param(
        svg_snippet(width="5mm", height="6mm", viewbox="0 0 100 100"),
        (resolve_length("5mm"), resolve_length("6mm")),
        Transform.scaling(resolve_length("5mm") / 100).translate(
            x=0, y=10 * resolve_length("5mm") / 100
        ),
        no_error(),
        id="fixed size, viewbox 100",
    ),
    pytest.param(
        svg_snippet(viewbox="0 0 110 90"),
        (10, 10),
        Transform.scaling(10 / 110).translate(x=0, y=10 * 10 / 110),
        no_error(),
        id="viewbox wide",
    ),
    pytest.param(
        svg_snippet(viewbox="0 0 90 110"),
        (10, 10),
        Transform.scaling(10 / 110).translate(x=10 * 10 / 110, y=0),
        no_error(),
        id="viewbox tall",
    ),
    pytest.param(
        svg_snippet(viewbox="0 0 110 90", aspect=False),
        (10, 10),
        Transform.scaling(10 / 110, 10 / 90),
        no_error(),
        id="viewbox wide, no aspect preservation",
    ),
    pytest.param(
        svg_snippet(viewbox="0 0 90 110", aspect=False),
        (10, 10),
        Transform.scaling(10 / 90, 10 / 110),
        no_error(),
        id="viewbox tall, no aspect preservation",
    ),
    pytest.param(
        svg_snippet(viewbox="45 55 90 110"),
        (10, 10),
        Transform.scaling(10 / 110, 10 / 110).translate(-35 / 11, -5),
        no_error(),
        id="viewbox tall and shifted",
    ),
    pytest.param(
        svg_snippet(viewbox="45 55 90 110", aspect=False),
        (10, 10),
        Transform.scaling(10 / 90, 10 / 110).translate(-5, -5),
        no_error(),
        id="viewbox tall and shifted, no aspect preservation",
    ),
    pytest.param(
        svg_snippet(viewbox="0 0 -90 110", aspect=False),
        (10, 10),
        None,
        pytest.raises(ValueError),
        id="invalid viewbox",
    ),
)

svg_current_color = (
    # if there's no color defined in the graphics context, we expect black
    # (that's the behaviour when the SVG is opened in a browser)
    pytest.param(
        None,
        None,
        "_black_by_default",
        id="red draw and",
    ),
    pytest.param(
        (255, 0, 0),
        (255, 0, 0),
        "_red_fill_draw",
        id="red fill and draw colors",
    ),
)



# pylint: disable=protected-access
from fpdf import FPDF
from fpdf.line_break import Fragment

PDF = FPDF()
GSTATE = PDF._get_current_graphics_state()
GSTATE_B = GSTATE.copy()
GSTATE_B.font_style = "B"
GSTATE_I = GSTATE.copy()
GSTATE_I.font_style = "I"
GSTATE_S = GSTATE.copy()
GSTATE_S.strikethrough = True
GSTATE_U = GSTATE.copy()
GSTATE_U.underline = True
GSTATE_BI = GSTATE.copy()
GSTATE_BI.font_style = "BI"


def merge_fragments(fragments):
    """
    Helper function for testing the escaping characters
    Will merge fragments that have different characters but same fragment.graphics_state
    and same fragment.k and same fragment.link.
    Example Input:
    (
    Fragment(characters=['a'], graphics_state={000}, k=1, link=None),
    Fragment(characters=['b'], graphics_state={000}, k=1, link=None)
    )
    Example Output:
    (Fragment(characters=['a', 'b'], graphics_state={000}, k=1, link=None))
    """
    if not fragments:
        return []

    merged_fragments = []
    current_fragment = fragments[0]

    for fragment in fragments[1:]:
        if (
            fragment.graphics_state == current_fragment.graphics_state
            and fragment.k == current_fragment.k
            and fragment.link == current_fragment.link
        ):
            current_fragment.characters.extend(fragment.characters)
        else:
            merged_fragments.append(current_fragment)
            current_fragment = fragment

    merged_fragments.append(current_fragment)

    return tuple(merged_fragments)


def test_markdown_parse_simple_ok():
    frags = tuple(
        FPDF()._parse_chars(
            "**bold**, __italics__, ~~strikethrough~~ and --underlined--", True
        )
    )
    expected = (
        Fragment("bold", GSTATE_B, k=PDF.k),
        Fragment(", ", GSTATE, k=PDF.k),
        Fragment("italics", GSTATE_I, k=PDF.k),
        Fragment(", ", GSTATE, k=PDF.k),
        Fragment("strikethrough", GSTATE_S, k=PDF.k),
        Fragment(" and ", GSTATE, k=PDF.k),
        Fragment("underlined", GSTATE_U, k=PDF.k),
    )
    assert frags == expected


def test_markdown_parse_simple_ok_escaped():
    frags = merge_fragments(
        tuple(
            FPDF()._parse_chars(
                "\\**bold\\**, \\__italics\\__ and \\--underlined\\-- escaped", True
            )
        )
    )
    expected = (
        Fragment("**bold**, __italics__ and --underlined-- escaped", GSTATE, k=PDF.k),
    )
    assert frags == expected
    frags = merge_fragments(
        tuple(
            FPDF()._parse_chars(
                r"raw \**bold\**, \__italics\__ and \--underlined\-- escaped", True
            )
        )
    )
    expected = (
        Fragment(
            "raw **bold**, __italics__ and --underlined-- escaped", GSTATE, k=PDF.k
        ),
    )
    assert frags == expected
    frags = tuple(FPDF()._parse_chars("escape *\\*between marker*\\*", True))
    expected = (Fragment("escape *\\*between marker*\\*", GSTATE, k=PDF.k),)
    assert frags == expected
    frags = tuple(FPDF()._parse_chars("escape **\\after marker**\\", True))
    expected = (
        Fragment("escape ", GSTATE, k=PDF.k),
        Fragment("\\after marker", GSTATE_B, k=PDF.k),
        Fragment("\\", GSTATE, k=PDF.k),
    )
    assert frags == expected


def test_markdown_parse_strikethrough_escaped():
    frags = merge_fragments(tuple(FPDF()._parse_chars("\\~~strike\\~~", True)))
    expected = (Fragment("~~strike~~", GSTATE, k=PDF.k),)
    assert frags == expected


def test_markdown_unrelated_escape():
    frags = merge_fragments(
        tuple(FPDF()._parse_chars("unrelated \\ escape \\**bold\\**", True))
    )
    expected = (Fragment("unrelated \\ escape **bold**", GSTATE, k=PDF.k),)
    assert frags == expected
    frags = merge_fragments(
        tuple(FPDF()._parse_chars("unrelated \\\\ double escape \\**bold\\**", True))
    )
    expected = (Fragment("unrelated \\\\ double escape **bold**", GSTATE, k=PDF.k),)
    assert frags == expected


def test_markdown_parse_multiple_escape():
    frags = merge_fragments(
        tuple(FPDF()._parse_chars("\\\\**bold\\\\** double escaped", True))
    )
    expected = (
        Fragment("\\\\", GSTATE, k=PDF.k),
        Fragment("bold\\\\", GSTATE_B, k=PDF.k),
        Fragment(" double escaped", GSTATE, k=PDF.k),
    )

    assert frags == expected
    frags = merge_fragments(
        tuple(FPDF()._parse_chars("\\\\\\**triple bold\\\\\\** escaped", True))
    )
    expected = (Fragment("\\\\**triple bold\\\\** escaped", GSTATE, k=PDF.k),)
    assert frags == expected


def test_markdown_parse_overlapping():
    frags = tuple(FPDF()._parse_chars("**bold __italics__**", True))
    expected = (
        Fragment("bold ", GSTATE_B, k=PDF.k),
        Fragment("italics", GSTATE_BI, k=PDF.k),
    )
    assert frags == expected


def test_markdown_parse_overlapping_escaped():
    frags = merge_fragments(
        tuple(FPDF()._parse_chars("**bold \\__italics\\__**", True))
    )
    expected = (Fragment("bold __italics__", GSTATE_B, k=PDF.k),)
    assert frags == expected


def test_markdown_parse_crossing_markers():
    frags = merge_fragments(
        tuple(FPDF()._parse_chars("**bold __and** italics__", True))
    )
    expected = (
        Fragment("bold ", GSTATE_B, k=PDF.k),
        Fragment("and", GSTATE_BI, k=PDF.k),
        Fragment(" italics", GSTATE_I, k=PDF.k),
    )
    assert frags == expected


def test_markdown_parse_crossing_markers_escaped():
    frags = merge_fragments(
        tuple(FPDF()._parse_chars("**bold __and\\** italics__", True))
    )
    expected = (
        Fragment("bold ", GSTATE_B, k=PDF.k),
        Fragment("and** italics", GSTATE_BI, k=PDF.k),
    )
    assert frags == expected


def test_markdown_parse_unterminated():
    frags = merge_fragments(tuple(FPDF()._parse_chars("**bold __italics__", True)))
    expected = (
        Fragment("bold ", GSTATE_B, k=PDF.k),
        Fragment("italics", GSTATE_BI, k=PDF.k),
    )
    assert frags == expected


def test_markdown_parse_unterminated_escaped():
    frags = merge_fragments(tuple(FPDF()._parse_chars("**bold\\** __italics__", True)))
    expected = (
        Fragment("bold** ", GSTATE_B, k=PDF.k),
        Fragment("italics", GSTATE_BI, k=PDF.k),
    )
    assert frags == expected


def test_markdown_parse_line_of_markers():
    frags = merge_fragments(tuple(FPDF()._parse_chars("*** woops", True)))
    expected = (Fragment("*** woops", GSTATE, k=PDF.k),)
    assert frags == expected
    frags = merge_fragments(tuple(FPDF()._parse_chars("----------", True)))
    expected = (Fragment("----------", GSTATE, k=PDF.k),)
    assert frags == expected
    frags = tuple(FPDF()._parse_chars("****BOLD**", True))
    expected = (Fragment("****BOLD", GSTATE, k=PDF.k),)
    assert frags == expected
    frags = merge_fragments(tuple(FPDF()._parse_chars("\\****BOLD**\\**", True)))
    expected = (
        Fragment("**", GSTATE, k=PDF.k),
        Fragment("BOLD", GSTATE_B, k=PDF.k),
        Fragment("**", GSTATE, k=PDF.k),
    )
    assert frags == expected
    frags = merge_fragments(tuple(FPDF()._parse_chars("* **BOLD**", True)))
    expected = (
        Fragment("* ", GSTATE, k=PDF.k),
        Fragment("BOLD", GSTATE_B, k=PDF.k),
    )
    assert frags == expected


def test_markdown_parse_line_of_markers_escaped():
    frags = merge_fragments(tuple(FPDF()._parse_chars("\\****BOLD**", True)))
    expected = (
        Fragment("**", GSTATE, k=PDF.k),
        Fragment("BOLD", GSTATE_B, k=PDF.k),
    )
    assert frags == expected
    frags = merge_fragments(tuple(FPDF()._parse_chars("*\\***BOLD**", True)))
    expected = (Fragment("****BOLD", GSTATE, k=PDF.k),)
    assert frags == expected


def test_markdown_parse_newline_after_markdown_link():  # issue 916
    text = "[fpdf2](https://py-pdf.github.io/fpdf2/)\nGo visit it!"
    frags = merge_fragments(tuple(FPDF()._parse_chars(text, True)))
    gstate_link = GSTATE.copy()
    gstate_link.underline = True
    expected = (
        Fragment(
            "fpdf2",
            gstate_link,
            k=PDF.k,
            link="https://py-pdf.github.io/fpdf2/",
        ),
        Fragment("\nGo visit it!", GSTATE, k=PDF.k),
    )
    assert frags == expected


def test_markdown_parse_trailing_escape():
    frags = merge_fragments(tuple(FPDF()._parse_chars("trailing \\\\", True)))
    expected = (Fragment("trailing \\\\", GSTATE, k=PDF.k),)
    assert frags == expected


def test_markdown_parse_escape_non_marker():
    frags = merge_fragments(tuple(FPDF()._parse_chars(r"\a", True)))
    expected = (Fragment(r"\a", GSTATE, k=PDF.k),)
    assert frags == expected


def test_markdown_parse_escape_before_marker_odd_even():
    frags = tuple(FPDF()._parse_chars("\\\\**bold**", True))
    expected = (
        Fragment("\\\\", GSTATE, k=PDF.k),
        Fragment("bold", GSTATE_B, k=PDF.k),
    )
    assert frags == expected

    frags = tuple(FPDF()._parse_chars("\\**bold**", True))
    expected = (
        Fragment("**", GSTATE, k=PDF.k),
        Fragment("bold", GSTATE, k=PDF.k),
    )
    assert frags == expected


def test_markdown_parse_marker_adjacency():
    frags = merge_fragments(tuple(FPDF()._parse_chars("a***b", True)))
    expected = (Fragment("a***b", GSTATE, k=PDF.k),)
    assert frags == expected

    frags = tuple(FPDF()._parse_chars("**bold***", True))
    expected = (Fragment("bold***", GSTATE_B, k=PDF.k),)
    assert frags == expected


def test_markdown_parse_across_newline():
    frags = tuple(FPDF()._parse_chars("**bold\nstill**", True))
    expected = (Fragment("bold\nstill", GSTATE_B, k=PDF.k),)
    assert frags == expected


def test_markdown_parse_nested_combinations():
    frags = tuple(FPDF()._parse_chars("**bold --under--**", True))
    gstate_bu = GSTATE.copy()
    gstate_bu.font_style = "B"
    gstate_bu.underline = True
    expected = (
        Fragment("bold ", GSTATE_B, k=PDF.k),
        Fragment("under", gstate_bu, k=PDF.k),
    )
    assert frags == expected

    frags = tuple(FPDF()._parse_chars("~~strike __ital__~~", True))
    gstate_si = GSTATE_S.copy()
    gstate_si.font_style = "I"
    expected = (
        Fragment("strike ", GSTATE_S, k=PDF.k),
        Fragment("ital", gstate_si, k=PDF.k),
    )
    assert frags == expected


def test_markdown_parse_link_variations():
    frags = tuple(FPDF()._parse_chars("[go](2)", True))
    assert len(frags) == 1
    assert "".join(frags[0].characters) == "go"
    assert isinstance(frags[0].link, int)
    assert frags[0].graphics_state.underline is True

    frags = tuple(FPDF()._parse_chars("[**bold**](https://example.com)", True))
    assert len(frags) == 1
    assert "".join(frags[0].characters) == "**bold**"
    assert frags[0].graphics_state.underline is True
    assert frags[0].graphics_state.font_style == ""
    assert frags[0].link == "https://example.com"

    frags = tuple(FPDF()._parse_chars("[x](url)**y**", True))
    assert len(frags) == 2
    assert "".join(frags[0].characters) == "x"
    assert frags[0].graphics_state.underline is True
    assert frags[0].link == "url"
    assert "".join(frags[1].characters) == "y"
    assert frags[1].graphics_state.font_style == "B"

    frags = tuple(FPDF()._parse_chars("\\[x](url)", True))
    expected = (
        Fragment("\\", GSTATE, k=PDF.k),
        Fragment("x", GSTATE_U, k=PDF.k, link="url"),
    )
    assert frags == expected
    assert frags[1].link == "url"



from pathlib import Path

import fpdf
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def draw_points(pdf, point_lists):
    for pl in point_lists:
        for p in pl:
            pdf.circle(x=p[0] + 0.5, y=p[1] + 0.5, radius=0.5, style="FD")


def test_quadratic_beziers(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    pl1 = [(20, 40), (40, 20), (60, 40)]
    pl2 = [(20, 80), (50, 100), (60, 80)]
    pl3 = [(20, 120), (40, 110), (60, 120)]
    pl4 = [(20, 170), (40, 160), (60, 170)]
    pl5 = [[20, 230], (40, 280), (60, 250)]

    pdf.set_fill_color(r=255, g=0, b=0)
    pdf.bezier(pl1, style="DF")
    pdf.set_fill_color(r=0, g=255, b=0)
    pdf.bezier(pl2, style="DF")
    pdf.set_fill_color(r=0, g=0, b=255)
    pdf.bezier(pl3, style="DF", closed=True)
    pdf.bezier(pl4, style="F")
    pdf.bezier(pl5, style="D")

    draw_points(pdf, [pl1, pl2, pl3, pl4, pl5])

    assert_pdf_equal(pdf, HERE / "quadratic_bezier_curve.pdf", tmp_path)


def test_cubic_beziers(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    pl1 = [(120, 40), (140, 30), (160, 49), (180, 50)]
    pl2 = [(120, 80), (150, 100), (160, 80), (180, 80)]
    pl3 = [(120, 120), (140, 130), (160, 140), (180, 120)]
    pl4 = [(20, 20), (40, 10), (60, 20)]
    pl5 = [[20, 80], (40, 90), (60, 80)]

    pdf.set_fill_color(r=255, g=0, b=0)
    pdf.bezier(pl1, style="DF")
    pdf.set_fill_color(r=0, g=255, b=0)
    pdf.bezier(pl2, style="DF")
    pdf.set_fill_color(r=0, g=0, b=255)
    pdf.bezier(pl3, style="DF", closed=True)
    pdf.bezier(pl4, style="F")
    pdf.bezier(pl5, style="D")

    draw_points(pdf, [pl1, pl2, pl3, pl4, pl5])

    assert_pdf_equal(pdf, HERE / "cubic_bezier_curve.pdf", tmp_path)


def test_bezier_line_settings(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    pl1 = [(120, 40), (140, 30), (160, 49), (180, 50)]
    pl2 = [(20, 80), (50, 100), (60, 80)]

    pdf.set_fill_color(r=255, g=0, b=0)
    pdf.set_dash_pattern(dash=2, gap=3)
    pdf.bezier(pl1, style="DF")

    pdf.set_fill_color(r=0, g=255, b=0)
    pdf.set_dash_pattern(dash=4, gap=6)
    pdf.set_line_width(2)
    pdf.bezier(pl2, style="DF")

    # Reset for drawing points
    pdf.set_line_width(0.2)
    pdf.set_dash_pattern(0, 0, 0)

    draw_points(pdf, [pl1, pl2])

    assert_pdf_equal(pdf, HERE / "bezier_curve_line_settings.pdf", tmp_path)


def test_bezier_chaining(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=12)

    pdf.x, pdf.y = 20, pdf.h / 4 - 30
    pdf.cell(text="Chaining cubic curves:")
    for i in range(4):
        shift = 20 if i % 2 else -20
        x1, y1 = (3 * i + 1) * pdf.w / 14, pdf.h / 4
        x2, y2 = (3 * i + 2) * pdf.w / 14, pdf.h / 4 + shift
        x3, y3 = (3 * i + 3) * pdf.w / 14, pdf.h / 4 + shift
        x4, y4 = (3 * i + 4) * pdf.w / 14, pdf.h / 4
        pdf.set_draw_color(
            {
                0: "#000",
                1: "#f00",
                2: "#0f0",
                3: "#00f",
            }[i % 4]
        )
        pdf.circle(x=x1, y=y1, radius=0.5)
        pdf.circle(x=x2, y=y2, radius=0.5)
        pdf.circle(x=x3, y=y3, radius=0.5)
        pdf.circle(x=x4, y=y4, radius=0.5)
        pdf.bezier(((x1, y1), (x2, y2), (x3, y3), (x4, y4)))

    pdf.x, pdf.y = 20, pdf.h / 2 - 30
    pdf.cell(text="Chaining quadratic curves:")
    for i in range(4):
        shift = 20 if i % 2 else -20
        x1, y1 = (2 * i + 1) * pdf.w / 10, pdf.h / 2
        x2, y2 = (2 * i + 2) * pdf.w / 10, pdf.h / 2 + shift
        x3, y3 = (2 * i + 3) * pdf.w / 10, pdf.h / 2
        pdf.set_draw_color(
            {
                0: "#f00",
                1: "#0f0",
                2: "#00f",
            }[i % 3]
        )
        pdf.circle(x=x1, y=y1, radius=0.5)
        pdf.circle(x=x2, y=y2, radius=0.5)
        pdf.circle(x=x3, y=y3, radius=0.5)
        pdf.bezier(((x1, y1), (x2, y2), (x3, y3)))

    assert_pdf_equal(pdf, HERE / "bezier_chaining.pdf", tmp_path)



"""
PDF template helpers for fpdf2.

Usage documentation at: <https://py-pdf.github.io/fpdf2/Templates.html>
"""

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2010 Mariano Reingart"
__license__ = "LGPL 3.0"

import csv
import json
import locale
import os
import warnings
from typing import Any, Optional, Sequence

from .deprecation import get_stack_level
from .enums import WrapMode
from .errors import FPDFException
from .fpdf import FPDF


def _rgb(col: int) -> tuple[float, float, float]:
    return (col // 65536), (col // 256 % 256), (col % 256)


def _rgb_as_str(col: int) -> str:
    r, g, b = _rgb(col)
    if (r == 0 and g == 0 and b == 0) or g == -1:
        return f"{r / 255:.3f} g"
    return f"{r / 255:.3f} {g / 255:.3f} {b / 255:.3f} rg"


class FlexTemplate:
    """
    A flexible templating class.

    Allows to apply one or several template definitions to any page of
    a document in any combination.

    Usage documentation at: <https://py-pdf.github.io/fpdf2/Templates.html>
    """

    def __init__(
        self, pdf: FPDF, elements: Optional[Sequence[dict[str, Any]]] = None
    ) -> None:
        """
        Args:
            pdf (fpdf.fpdf.FPDF): All content will be added to this `FPDF` instance.
            elements (list[dict]): A template definition in a list of dicts.
                If you omit this, then you need to call either load_elements()
                or parse_csv() before doing anything else.
        """
        if not isinstance(pdf, FPDF):
            raise TypeError("'pdf' must be an instance of fpdf.FPDF()")
        self.pdf = pdf
        self.splitting_pdf: Optional[FPDF] = None  # for split_multicell()
        if elements:
            self.load_elements(elements)
        self.handlers = {
            "T": self._text,
            "L": self._line,
            "I": self._image,
            "B": self._rect,
            "E": self._ellipse,
            "BC": self._barcode,
            "C39": self._code39,
            "W": self._write,
        }
        self.texts: dict[str, Any] = {}

    def load_elements(self, elements: Sequence[dict[str, Any]]) -> None:
        """
        Load a template definition.

        Args:
            elements (list of dicts): A template definition in a list of dicts
        """
        key_config = {
            # key: type
            "name": (str, type(None)),
            "type": (str, type(None)),
            "x1": (int, float),
            "y1": (int, float),
            "x2": (int, float),
            "y2": (int, float),
            "font": (str, type(None)),
            "size": (int, float),
            "bold": object,  # "bool or equivalent"
            "italic": object,
            "underline": object,
            "foreground": int,
            "background": int,
            "align": (str, type(None)),
            "text": (str, type(None)),
            "priority": int,
            "multiline": (bool, type(None)),
            "rotate": (int, float),
            "wrapmode": (str, type(None)),
            "dash_pattern": (dict, type(None)),
        }

        self.elements: Sequence[dict[str, Any]] = elements
        self.keys: list[str] = []
        for e in elements:
            # priority is optional, but we need a default for sorting.
            if not "priority" in e:
                e["priority"] = 0
            for k in ("name", "type", "x1", "y1", "y2"):
                if k not in e:
                    if e["type"] == "C39":
                        # lots of legacy special casing.
                        # We need to do that here, so that rotation and scaling
                        # still work.
                        if k == "x1" and "x" in e:
                            e["x1"] = e["x"]
                            continue
                        if k == "y1" and "y" in e:
                            e["y1"] = e["y"]
                            continue
                        if k == "y2" and "h" in e:
                            e["y2"] = e["y1"] + e["h"]
                            continue
                    raise KeyError(f"Mandatory key '{k}' missing in input data")
            # x2 is optional for barcode types, but needed for offset rendering
            if "x2" not in e:
                if e["type"] in ["BC", "C39"]:
                    e["x2"] = 0
                else:
                    raise KeyError("Mandatory key 'x2' missing in input data")
            if not "size" in e and e["type"] == "C39":
                if "w" in e:
                    e["size"] = e["w"]
            if "dash_pattern" in e:
                for k in e["dash_pattern"].keys():
                    if k not in ("dash", "gap", "phase"):
                        raise KeyError(f"Invalid key '{k}' in dash_pattern")
                if "dash" not in e["dash_pattern"]:
                    e["dash_pattern"]["dash"] = 0
                if "gap" not in e["dash_pattern"]:
                    e["dash_pattern"]["gap"] = e["dash_pattern"]["dash"]
                if "phase" not in e["dash_pattern"]:
                    e["dash_pattern"]["phase"] = 0

            for k, t in key_config.items():
                if k in e and not isinstance(e[k], t):  # type: ignore[arg-type]
                    ttype = (
                        t.__name__
                        if isinstance(t, type)
                        else " or ".join([f"'{x.__name__}'" for x in t])  # type: ignore[attr-defined]
                    )
                    raise TypeError(
                        f"Value of element item '{k}' must be {ttype}, not '{type(e[k]).__name__}'."
                    )
            self.keys.append(e["name"].lower())

    @staticmethod
    def _parse_colorcode(s: str) -> int:
        """Allow hex and oct values for colors"""
        if s[:2] in ["0x", "0X"]:
            return int(s, 16)
        if s[0] == "0":
            return int(s, 8)
        return int(s)

    @staticmethod
    def _parse_multiline(s: str) -> Optional[bool]:
        i = int(s)
        if i > 0:
            return True
        if i < 0:
            return False
        return None

    def parse_json(self, infile: os.PathLike[str], encoding: str = "utf-8") -> None:
        """
        Load the template definition from a JSON file.
        The data must be structured as an array of objects, with names and values exactly
        equivalent to what would get supplied to load_elements(),

        Args:
            infile (str, path-like object): The filepath of the JSON file.
            encoding (str): The character encoding of the file. Default is UTF-8.
        """
        with open(infile, encoding=encoding) as f:
            data = json.load(f)
            for d in data:
                fgval = d.get("foreground")
                if fgval and isinstance(fgval, str):
                    if fgval.lower().startswith("#"):
                        d["foreground"] = int(fgval[1:], 16)
                    else:
                        raise ValueError(
                            "If foreground is a string, it must have the form '#rrggbb'."
                        )
                bgval = d.get("background")
                if bgval and isinstance(bgval, str):
                    if bgval.lower().startswith("#"):
                        d["background"] = int(bgval[1:], 16)
                    else:
                        raise ValueError(
                            "If background is a string, it must have the form '#rrggbb'."
                        )
            self.load_elements(data)

    def parse_csv(
        self,
        infile: os.PathLike[str],
        delimiter: str = ",",
        decimal_sep: str = ".",
        encoding: Optional[str] = None,
    ) -> None:
        """
        Load the template definition from a CSV file.

        Args:
            infile (str, path-like object): The filepath of the CSV file.
            delimiter (str): The character that separates the fields in the CSV file:
                Usually a comma, semicolon, or tab.
            decimal_sep (str): The decimal separator used in the file.
                Usually either a point or a comma.
            encoding (str): The character encoding of the file.
                Default is the system default encoding.
        """

        def _varsep_float(s: str, default: str = "0") -> float:
            """Convert to float with given decimal separator"""
            # glad to have nonlocal scoping...
            return float((s.strip() or default).replace(decimal_sep, "."))

        key_config = (
            # key, converter, mandatory
            ("name", str, True),
            ("type", str, True),
            ("x1", _varsep_float, True),
            ("y1", _varsep_float, True),
            ("x2", _varsep_float, True),
            ("y2", _varsep_float, True),
            ("font", str, False),
            ("size", _varsep_float, False),
            ("bold", int, False),
            ("italic", int, False),
            ("underline", int, False),
            ("foreground", self._parse_colorcode, False),
            ("background", self._parse_colorcode, False),
            ("align", str, False),
            ("text", str, False),
            ("priority", int, False),
            ("multiline", self._parse_multiline, False),
            ("rotate", _varsep_float, False),
            ("wrapmode", str, False),
        )
        self.elements = []
        if encoding is None:
            encoding = locale.getpreferredencoding()
        with open(infile, encoding=encoding) as f:
            for row in csv.reader(f, delimiter=delimiter):
                # fill in blanks for any missing items
                row.extend([""] * (len(key_config) - len(row)))
                kargs: dict[str, Any] = {}
                for val, cfg in zip(row, key_config):
                    vs = val.strip()
                    if not vs:
                        if cfg[2]:  # mandatory
                            if cfg[0] == "x2" and row[1] in ["BC", "C39"]:
                                # two types don't need x2, but offset rendering does
                                pass
                            else:
                                raise FPDFException(
                                    f"Mandatory value '{cfg[0]}' missing in csv data"
                                )
                        elif cfg[0] == "priority":
                            # formally optional, but we need some value for sorting
                            kargs["priority"] = 0
                        # otherwise, let the type handlers use their own defaults
                    else:
                        kargs[cfg[0]] = cfg[1](vs)  # type: ignore[operator]
                self.elements.append(kargs)
        self.keys = [val["name"].lower() for val in self.elements]

    def __setitem__(self, name: str, value: Any) -> None:
        assert isinstance(
            name, str
        ), f"name must be of type 'str', not '{type(name).__name__}'."
        # value has too many valid types to reasonably check here
        if name.lower() not in self.keys:
            raise FPDFException(f"Element not loaded, cannot set item: {name}")
        self.texts[name.lower()] = value

    # setitem shortcut (may be further extended)
    set = __setitem__

    def __contains__(self, name: str) -> bool:
        assert isinstance(
            name, str
        ), f"name must be of type 'str', not '{type(name).__name__}'."
        return name.lower() in self.keys

    def __getitem__(self, name: str) -> Any:
        assert isinstance(
            name, str
        ), f"name must be of type 'str', not '{type(name).__name__}'."
        if name not in self.keys:
            raise KeyError(name)
        key = name.lower()
        if key in self.texts:
            # text for this page:
            return self.texts[key]
        # find first element for default text:
        return next(
            (x["text"] for x in self.elements if x["name"].lower() == key), None
        )

    def split_multicell(self, text: str, element_name: str) -> list[str]:
        """
        Split a string between words, for the parts to fit into a given element
        width. Additional splits will be made replacing any '\\n' characters.

        Args:
            text (str): The input text string.
            element_name (str): The name of the template element to fit the text inside.

        Returns:
            A list of substrings, each of which will fit into the element width
            when rendered in the element font style and size.
        """
        element = next(
            element
            for element in self.elements
            if element["name"].lower() == element_name.lower()
        )
        if not self.splitting_pdf:
            self.splitting_pdf = FPDF()
            self.splitting_pdf.add_page()
        assert self.splitting_pdf is not None
        style = ""
        if element.get("bold"):
            style += "B"
        if element.get("italic"):
            style += "I"
        if element.get("underline"):
            style += "U"
        self.splitting_pdf.set_font(element["font"], style, element["size"])
        return self.splitting_pdf.multi_cell(  # type: ignore[return-value]
            w=element["x2"] - element["x1"],
            h=element["y2"] - element["y1"],
            text=str(text),
            align=element.get("align", ""),
            dry_run=True,
            output="LINES",
            wrapmode=element.get("wrapmode", "WORD"),
        )

    def _text(
        self,
        *_: Any,
        x1: float = 0,
        y1: float = 0,
        x2: float = 0,
        y2: float = 0,
        text: str = "",
        font: str = "helvetica",
        size: float = 10,
        scale: float = 1.0,
        bold: bool = False,
        italic: bool = False,
        underline: bool = False,
        align: str = "",
        foreground: int = 0,
        background: Optional[int] = None,
        multiline: Optional[bool] = None,
        wrapmode: str = "WORD",
        **__: Any,
    ) -> None:
        if not text:
            return
        pdf = self.pdf
        if pdf.text_color != _rgb_as_str(foreground):  # type: ignore[comparison-overlap]
            pdf.set_text_color(*_rgb(foreground))
        if background is None:
            fill = False
        else:
            fill = True
            if pdf.fill_color != _rgb_as_str(background):  # type: ignore[comparison-overlap]
                pdf.set_fill_color(*_rgb(background))

        font = font.strip().lower()
        style = ""
        for tag in "B", "I", "U":
            if text.startswith(f"<{tag}>") and text.endswith(f"</{tag}>"):
                text = text[3:-4]
                style += tag
        if bold:
            style += "B"
        if italic:
            style += "I"
        if underline:
            style += "U"
        pdf.set_font(font, style, size * scale)
        pdf.set_xy(x1, y1)
        width, height = x2 - x1, y2 - y1
        if multiline is None:  # write without wrapping/trimming (default)
            pdf.cell(w=width, h=height, text=text, border=0, align=align, fill=fill)
        elif multiline:  # automatic word - warp
            pdf.multi_cell(
                w=width,
                h=height,
                text=text,
                border=0,
                align=align,
                fill=fill,
                wrapmode=WrapMode.coerce(wrapmode),
            )
        else:  # trim to fit exactly the space defined
            text = pdf.multi_cell(  # type: ignore[index,assignment]
                w=width,
                h=height,
                text=text,
                align=align,
                wrapmode=WrapMode.coerce(wrapmode),
                dry_run=True,
                output="LINES",
            )[0]
            pdf.cell(w=width, h=height, text=text, border=0, align=align, fill=fill)

    def _line(
        self,
        *_: Any,
        x1: float = 0,
        y1: float = 0,
        x2: float = 0,
        y2: float = 0,
        size: float = 0,
        scale: float = 1.0,
        foreground: int = 0,
        dash_pattern: Optional[tuple[float, float, float]] = None,
        **__: Any,
    ) -> None:
        assert self.pdf.draw_color is not None
        if self.pdf.draw_color.serialize().lower() != _rgb_as_str(foreground):
            self.pdf.set_draw_color(*_rgb(foreground))
        self.pdf.set_line_width(size * scale)
        if dash_pattern is not None:
            # Save dash_pattern to restore after this line
            restore_dash_pattern = self.pdf.dash_pattern
            if isinstance(dash_pattern, dict):
                self.pdf.set_dash_pattern(
                    dash_pattern.get(  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                        "dash", 0
                    ),
                    dash_pattern.get(  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                        "gap", 0
                    ),
                    dash_pattern.get(  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                        "phase", 0
                    ),
                )
            else:
                self.pdf.set_dash_pattern(
                    dash_pattern[0], dash_pattern[1], dash_pattern[2]
                )
        self.pdf.line(x1, y1, x2, y2)
        if dash_pattern is not None:
            self.pdf.set_dash_pattern(
                **restore_dash_pattern  # pyright: ignore[reportPossiblyUnboundVariable]
            )

    def _rect(
        self,
        *_: Any,
        x1: float = 0,
        y1: float = 0,
        x2: float = 0,
        y2: float = 0,
        size: float = 0,
        scale: float = 1.0,
        foreground: int = 0,
        background: Optional[int] = None,
        dash_pattern: Optional[tuple[float, float, float]] = None,
        **__: Any,
    ) -> None:
        pdf = self.pdf
        assert pdf.draw_color is not None
        if pdf.draw_color.serialize().lower() != _rgb_as_str(foreground):
            pdf.set_draw_color(*_rgb(foreground))
        if background is None:
            style = "D"
        else:
            style = "FD"
            if pdf.fill_color != _rgb_as_str(background):  # type: ignore[comparison-overlap]
                pdf.set_fill_color(*_rgb(background))
        pdf.set_line_width(size * scale)
        if dash_pattern is not None:
            # Save dash_pattern to restore after this rect
            restore_dash_pattern = self.pdf.dash_pattern
            if isinstance(dash_pattern, dict):
                pdf.set_dash_pattern(
                    dash_pattern.get(  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                        "dash", 0
                    ),
                    dash_pattern.get(  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                        "gap", 0
                    ),
                    dash_pattern.get(  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                        "phase", 0
                    ),
                )
            else:
                pdf.set_dash_pattern(dash_pattern[0], dash_pattern[1], dash_pattern[2])
        pdf.rect(x1, y1, x2 - x1, y2 - y1, style=style)
        if dash_pattern is not None:
            self.pdf.set_dash_pattern(
                **restore_dash_pattern  # pyright: ignore[reportPossiblyUnboundVariable]
            )

    def _ellipse(
        self,
        *_: Any,
        x1: float = 0,
        y1: float = 0,
        x2: float = 0,
        y2: float = 0,
        size: float = 0,
        scale: float = 1.0,
        foreground: int = 0,
        background: Optional[int] = None,
        dash_pattern: Optional[tuple[float, float, float]] = None,
        **__: Any,
    ) -> None:
        pdf = self.pdf
        assert pdf.draw_color is not None
        if pdf.draw_color.serialize().lower() != _rgb_as_str(foreground):
            pdf.set_draw_color(*_rgb(foreground))
        if background is None:
            style = "D"
        else:
            style = "FD"
            if pdf.fill_color != _rgb_as_str(background):  # type: ignore[comparison-overlap]
                pdf.set_fill_color(*_rgb(background))
        pdf.set_line_width(size * scale)
        if dash_pattern is not None:
            # Save dash_pattern to restore after this ellipse
            restore_dash_pattern = self.pdf.dash_pattern
            if isinstance(dash_pattern, dict):
                pdf.set_dash_pattern(
                    dash_pattern.get(  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                        "dash", 0
                    ),
                    dash_pattern.get(  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                        "gap", 0
                    ),
                    dash_pattern.get(  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
                        "phase", 0
                    ),
                )
            else:
                pdf.set_dash_pattern(dash_pattern[0], dash_pattern[1], dash_pattern[2])
        pdf.ellipse(x1, y1, x2 - x1, y2 - y1, style=style)
        if dash_pattern is not None:
            self.pdf.set_dash_pattern(
                **restore_dash_pattern  # pyright: ignore[reportPossiblyUnboundVariable]
            )

    def _image(
        self,
        *_: Any,
        x1: float = 0,
        y1: float = 0,
        x2: float = 0,
        y2: float = 0,
        text: str = "",
        **__: Any,
    ) -> None:
        if text:
            self.pdf.image(text, x1, y1, w=x2 - x1, h=y2 - y1, link="")

    def _barcode(
        self,
        *_: Any,
        x1: float = 0,
        y1: float = 0,
        x2: float = 0,
        y2: float = 0,
        text: str = "",
        font: str = "interleaved 2of5 nt",
        size: float = 1,
        scale: float = 1.0,
        foreground: int = 0,
        **__: Any,
    ) -> None:
        # pylint: disable=unused-argument
        pdf = self.pdf
        assert pdf.fill_color is not None
        if pdf.fill_color.serialize().lower() != _rgb_as_str(foreground):
            pdf.set_fill_color(*_rgb(foreground))
        font = font.lower().strip()
        if font == "interleaved 2of5 nt":
            pdf.interleaved2of5(text, x1, y1, w=size * scale, h=y2 - y1)

    def _code39(
        self,
        *_: Any,
        x1: float = 0,
        y1: float = 0,
        y2: float = 0,
        text: str = "",
        size: float = 1.5,
        scale: float = 1.0,
        foreground: int = 0,
        x: Optional[float] = None,
        y: Optional[float] = None,
        w: Optional[float] = None,
        h: Optional[float] = None,
        **__: Any,
    ) -> None:
        if x is not None or y is not None or w is not None or h is not None:
            warnings.warn(
                (
                    "code39 arguments x/y/w/h are deprecated since v2.4.4,"
                    " please use x1/y1/y2/size instead"
                ),
                DeprecationWarning,
                stacklevel=get_stack_level(),
            )
        pdf = self.pdf
        assert pdf.fill_color is not None
        if pdf.fill_color.serialize().lower() != _rgb_as_str(foreground):
            pdf.set_fill_color(*_rgb(foreground))
        h = y2 - y1
        if h <= 0:
            h = 5
        pdf.code39(text, x1, y1, size * scale, h)

    # Added by Derek Schwalenberg Schwalenberg1013@gmail.com to allow (url) links in
    # templates (using write method) 2014-02-22
    def _write(
        self,
        *_: Any,
        x1: float = 0,
        y1: float = 0,
        x2: float = 0,
        y2: float = 0,
        text: str = "",
        font: str = "helvetica",
        size: float = 10,
        scale: float = 1.0,
        bold: bool = False,
        italic: bool = False,
        underline: bool = False,
        link: Optional[int | str] = "",
        foreground: int = 0,
        **__: Any,
    ) -> None:
        # pylint: disable=unused-argument
        if not text:
            return
        pdf = self.pdf
        if pdf.text_color != _rgb_as_str(foreground):  # type: ignore[comparison-overlap]
            pdf.set_text_color(*_rgb(foreground))
        font = font.strip().lower()
        style = ""
        for tag in "B", "I", "U":
            if text.startswith(f"<{tag}>") and text.endswith(f"</{tag}>"):
                text = text[3:-4]
                style += tag
        if bold:
            style += "B"
        if italic:
            style += "I"
        if underline:
            style += "U"
        pdf.set_font(font, style, size * scale)
        pdf.set_xy(x1, y1)
        pdf.write(5, text, link)

    def render(
        self,
        offsetx: float = 0.0,
        offsety: float = 0.0,
        rotate: float = 0.0,
        scale: float = 1.0,
    ) -> None:
        """
        Add the contents of the template to the PDF document.

        Args:
            offsetx, offsety (float): Place the template to move its origin to the given coordinates.
            rotate (float): Rotate the inserted template around its (offset) origin.
            scale (float): Scale the inserted template by this factor.
        """
        sorted_elements = sorted(self.elements, key=lambda x: x["priority"])
        with self.pdf.local_context():
            for element in sorted_elements:
                ele = element.copy()  # don't want to modify the callers original
                ele["text"] = self.texts.get(ele["name"].lower(), ele.get("text", ""))
                if scale != 1.0:
                    ele["x1"] = ele["x1"] * scale
                    ele["y1"] = ele["y1"] * scale
                    ele["x2"] = ele["x1"] + ((ele["x2"] - element["x1"]) * scale)
                    ele["y2"] = ele["y1"] + ((ele["y2"] - element["y1"]) * scale)
                if offsetx:
                    ele["x1"] = ele["x1"] + offsetx
                    ele["x2"] = ele["x2"] + offsetx
                if offsety:
                    ele["y1"] = ele["y1"] + offsety
                    ele["y2"] = ele["y2"] + offsety
                ele["scale"] = scale
                handler_name = ele["type"].upper()
                if rotate:  # don't rotate by 0.0 degrees
                    with self.pdf.rotation(rotate, offsetx, offsety):
                        if "rotate" in ele and ele["rotate"]:
                            with self.pdf.rotation(ele["rotate"], ele["x1"], ele["y1"]):
                                self.handlers[handler_name](**ele)  # type: ignore[operator]
                        else:
                            self.handlers[handler_name](**ele)  # type: ignore[operator]
                else:
                    if "rotate" in ele and ele["rotate"]:
                        with self.pdf.rotation(ele["rotate"], ele["x1"], ele["y1"]):
                            self.handlers[handler_name](**ele)  # type: ignore[operator]
                    else:
                        self.handlers[handler_name](**ele)  # type: ignore[operator]
        self.texts = {}  # reset modified entries for the next page


class Template(FlexTemplate):
    """
    A simple templating class.

    Allows to apply a single template definition to all pages of a document.

    Usage documentation at: <https://py-pdf.github.io/fpdf2/Templates.html>
    """

    # Disabling this check due to the "format" parameter below:
    # pylint: disable=redefined-builtin
    def __init__(
        self,
        infile: Optional[os.PathLike[str]] = None,
        elements: Optional[Sequence[dict[str, Any]]] = None,
        format: str = "A4",
        orientation: str = "portrait",
        unit: str = "mm",
        title: str = "",
        author: str = "",
        subject: str = "",
        creator: str = "",
        keywords: str = "",
    ) -> None:
        """
        Args:
            infile (str): [**DEPRECATED since 2.2.0**] unused, will be removed in a later version
            elements (list[dict]): A template definition in a list of dicts.
                If you omit this, then you need to call either load_elements()
                or parse_csv() before doing anything else.
            format (str): The page format of the document (eg. "A4" or "letter").
            orientation (str): The orientation of the document.
                Possible values are "portrait"/"P" or "landscape"/"L"
            unit (str): The units used in the template definition.
                One of "mm", "cm", "in", "pt", or a number for points per unit.
            title (str): The title of the document.
            author (str): The author of the document.
            subject (str): The subject matter of the document.
            creator (str): The creator of the document.
        """
        if infile:
            warnings.warn(
                '"infile" is deprecated since v2.2.0, unused and will soon be removed',
                DeprecationWarning,
                stacklevel=get_stack_level(),
            )
        for arg in (
            "format",
            "orientation",
            "unit",
            "title",
            "author",
            "subject",
            "creator",
            "keywords",
        ):
            # nosemgrep: python.lang.security.dangerous-globals-use.dangerous-globals-use
            if not isinstance(locals()[arg], str):
                raise TypeError(f'Argument "{arg}" must be of type str.')
        pdf = FPDF(format=format, orientation=orientation, unit=unit)
        pdf.set_title(title)
        pdf.set_author(author)
        pdf.set_creator(creator)
        pdf.set_subject(subject)
        pdf.set_keywords(keywords)
        super().__init__(pdf=pdf, elements=elements)

    def add_page(self) -> None:
        """Finish the current page, and proceed to the next one."""
        if self.pdf.page:
            self.render()
        self.pdf.add_page()

    # pylint: disable=arguments-differ
    def render(self, outfile: Optional[str] = None, dest: Optional[str] = None) -> None:  # type: ignore[override]
        """
        Finish the document and process all pending data.

        Args:
            outfile (str): If given, the PDF file will be written to this file name.
                Alternatively, the `.pdf.output()` method can be manually called.
            dest (str): [**DEPRECATED since 2.2.0**] unused, will be removed in a later version.
        """
        if dest:
            warnings.warn(
                '"dest" is deprecated since v2.2.0, unused and will soon be removed',
                DeprecationWarning,
                stacklevel=get_stack_level(),
            )
        self.pdf.set_font("helvetica", style="B", size=16)
        self.pdf.set_auto_page_break(False, margin=0)
        super().render()
        if outfile:
            self.pdf.output(outfile)



from io import BytesIO
from fpdf import FPDF
from barcode import Code128
from barcode.writer import SVGWriter

# Create a new PDF document
pdf = FPDF()
pdf.add_page()

# Set the position and size of the image in the PDF
x = 50
y = 50
w = 100
h = 70

# Generate a Code128 Barcode as SVG:
svg_img_bytes = BytesIO()
Code128("100000902922", writer=SVGWriter()).write(svg_img_bytes)
pdf.image(svg_img_bytes, x=x, y=y, w=w, h=h)

# Output a PDF file:
pdf.output("code128_barcode.pdf")



from pathlib import Path

from fpdf import FPDF
from fpdf.actions import GoToAction, GoToRemoteAction, LaunchAction, NamedAction
from fpdf.enums import AnnotationName
from fpdf.syntax import DestinationXYZ

from test.conftest import assert_pdf_equal, EPOCH, LOREM_IPSUM

HERE = Path(__file__).resolve().parent


def test_text_annotation(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=12)
    all_visible_flags = (
        "PRINT",
        "NO_ZOOM",
        "NO_ROTATE",
        "READ_ONLY",
        "LOCKED",
        "TOGGLE_NO_VIEW",
        "LOCKED_CONTENTS",
    )
    for i, flags in enumerate((("PRINT",), all_visible_flags)):
        for j, flag in enumerate(flags):
            pdf.text(x=15 + 50 * i, y=10 + 5 * j, text=flag)
        for j, name in enumerate(
            (None,)
            + tuple(AnnotationName.__members__.keys())
            + tuple(AnnotationName.__members__.values())
        ):
            pdf.text_annotation(
                x=20 + 50 * i,
                y=50 + 15 * j,
                text=f"This is a {name or 'default'} annotation.",
                name=name,
                flags=flags,
            )
    assert_pdf_equal(pdf, HERE / "text_annotation.pdf", tmp_path)


def test_named_actions(tmp_path):
    pdf = FPDF()
    pdf.set_font("Helvetica", size=24)
    pdf.add_page()
    pdf.text(x=80, y=140, text="First page")
    pdf.add_page()
    pdf.underline = True
    for x, y, named_action in (
        (40, 80, "NextPage"),
        (120, 80, "PrevPage"),
        (40, 200, "FirstPage"),
        (120, 200, "LastPage"),
    ):
        pdf.text(x=x, y=y, text=named_action)
        pdf.add_action(
            NamedAction(named_action),
            x=x,
            y=y - pdf.font_size,
            w=pdf.get_string_width(named_action),
            h=pdf.font_size,
        )
    pdf.underline = False
    pdf.add_page()
    pdf.text(x=80, y=140, text="Last page")
    assert_pdf_equal(pdf, HERE / "named_actions.pdf", tmp_path)


def test_goto_action(tmp_path):
    pdf = FPDF()
    pdf.set_font("Helvetica", size=24)
    pdf.add_page()
    x, y, text = 80, 140, "GoTo action"
    pdf.text(x=x, y=y, text=text)
    pdf.add_action(
        GoToAction(dest=DestinationXYZ(page=2, top=pdf.h_pt)),
        x=x,
        y=y - pdf.font_size,
        w=pdf.get_string_width(text),
        h=pdf.font_size,
    )
    pdf.add_page()
    pdf.text(x=80, y=140, text="Page 2")
    assert_pdf_equal(pdf, HERE / "goto_action.pdf", tmp_path)


def test_goto_action_named_destination(tmp_path):
    pdf = FPDF()
    pdf.set_font("Helvetica", size=24)
    pdf.add_page()
    x, y, text = 80, 140, "GoTo action"
    pdf.text(x=x, y=y, text=text)
    pdf.add_action(
        GoToAction(dest="#page2"),
        x=x,
        y=y - pdf.font_size,
        w=pdf.get_string_width(text),
        h=pdf.font_size,
    )
    pdf.add_page()
    pdf.add_link(y=140, name="page2")
    pdf.text(x=80, y=140, text="Page 2")
    assert_pdf_equal(pdf, HERE / "goto_action_named_destination.pdf", tmp_path)


def test_goto_remote_action(tmp_path):
    pdf = FPDF()
    pdf.set_font("Helvetica", size=24)
    pdf.add_page()
    x, y, text = 80, 140, "GoTo-Remote action"
    pdf.text(x=x, y=y, text=text)
    dest = DestinationXYZ(page=1, top=pdf.h_pt)
    pdf.add_action(
        GoToRemoteAction("goto_action.pdf", dest=dest),
        x=x,
        y=y - pdf.font_size,
        w=pdf.get_string_width(text),
        h=pdf.font_size,
    )
    assert_pdf_equal(pdf, HERE / "goto_remote_action.pdf", tmp_path)


def test_launch_action(tmp_path):
    pdf = FPDF()
    pdf.set_font("Helvetica", size=24)
    pdf.add_page()
    x, y, text = 80, 140, "Launch action"
    pdf.text(x=x, y=y, text=text)
    pdf.add_action(
        LaunchAction(file="goto_action.pdf"),
        x=x,
        y=y - pdf.font_size,
        w=pdf.get_string_width(text),
        h=pdf.font_size,
    )
    assert_pdf_equal(pdf, HERE / "launch_action.pdf", tmp_path)


def test_highlighted(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=24)
    with pdf.highlight("Highlight comment", type="Squiggly", modification_time=EPOCH):
        pdf.text(50, 50, "Line 1")
        pdf.set_y(50)
        pdf.multi_cell(w=30, text="Line 2")
    pdf.cell(w=60, text="Not highlighted", border=1)
    assert_pdf_equal(pdf, HERE / "highlighted.pdf", tmp_path)


def test_highlighted_over_page_break(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=24)
    pdf.write(text=LOREM_IPSUM)
    pdf.ln()
    with pdf.highlight("Comment", title="Freddy Mercury", modification_time=EPOCH):
        pdf.write(text=LOREM_IPSUM)
    assert_pdf_equal(pdf, HERE / "highlighted_over_page_break.pdf", tmp_path)


def test_ink_annotation(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=24)
    pdf.text(50, 50, "Some text")
    pdf.ink_annotation(
        [(40, 50), (70, 25), (100, 50), (70, 75), (40, 50)],
        title="Lucas",
        text="Hello world!",
    )
    assert_pdf_equal(pdf, HERE / "ink_annotation.pdf", tmp_path)


def test_free_text_annotation_all_parameters(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=24)
    pdf.text(x=50, y=50, text="Some text.")
    pdf.set_font_size(12)
    pdf.set_draw_color(255, 0, 0)
    pdf.free_text_annotation(
        x=100,
        y=130,
        text="This is a free text annotation.",
        w=150,
        h=15,
    )
    assert_pdf_equal(pdf, HERE / "free_text_annotation_all_parameters.pdf", tmp_path)


def test_free_text_annotation_text_parameter(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=24)
    pdf.text(x=50, y=50, text="Some text.")
    pdf.set_font_size(12)
    pdf.free_text_annotation(text="This is a free text annotation.")
    assert_pdf_equal(pdf, HERE / "free_text_annotation_text_parameter.pdf", tmp_path)


def test_free_text_annotation_width_parameter(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=24)
    pdf.text(x=50, y=50, text="Some text.")
    pdf.set_font_size(12)
    pdf.free_text_annotation(text="This is a free text annotation.", w=80)
    assert_pdf_equal(pdf, HERE / "free_text_annotation_width_parameter.pdf", tmp_path)



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal, EMOJI_TEST_TEXT

HERE = Path(__file__).resolve().parent
FONTS_DIR = HERE.parent / "fonts"


def test_noto_color_emoji(tmp_path):
    pdf = FPDF()
    pdf.add_font("NotoCBDT", "", HERE / "NotoColorEmoji-CBDT.ttf")
    pdf.add_page()
    test_text = "😂❤🤣👍😭🙏😘🥰😍😊"
    pdf.set_font("helvetica", "", 24)
    pdf.cell(text="Noto Color Emoji (CBDT)", new_x="lmargin", new_y="next")
    pdf.cell(text="Top 10 emojis:", new_x="right", new_y="top")
    pdf.set_font("NotoCBDT", "", 24)
    pdf.cell(text=test_text, new_x="lmargin", new_y="next")
    assert_pdf_equal(pdf, HERE / "cbdt_noto_color_emoji.pdf", tmp_path)


def test_noto_emoji_shaping(tmp_path):
    pdf = FPDF()
    pdf.add_font("NotoCBDT", "", HERE / "NotoColorEmoji-CBDT.ttf")
    pdf.add_page()
    combined_emojis = "🇫🇷 🇺🇸 🇨🇦 🧑 🧑🏽 🧑🏿"
    pdf.set_font("helvetica", "", 24)
    pdf.cell(text="Emojis without text shaping:", new_x="lmargin", new_y="next")
    pdf.set_font("NotoCBDT", "", 24)
    pdf.multi_cell(w=pdf.epw, text=combined_emojis, new_x="lmargin", new_y="next")
    pdf.ln()
    pdf.set_font("helvetica", "", 24)
    pdf.cell(text="Emojis with text shaping:", new_x="lmargin", new_y="next")
    pdf.set_font("NotoCBDT", "", 24)
    pdf.set_text_shaping(True)
    pdf.multi_cell(w=pdf.epw, text=combined_emojis, new_x="lmargin", new_y="next")
    assert_pdf_equal(pdf, HERE / "cbdt_noto_color_emoji_shaping.pdf", tmp_path)


def test_noto_emoji_text(tmp_path):
    text = EMOJI_TEST_TEXT
    pdf = FPDF()
    pdf.add_font("Roboto", "", FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("NotoCBDT", "", HERE / "NotoColorEmoji-CBDT.ttf")
    pdf.set_font("Roboto", "", 24)
    pdf.set_fallback_fonts(["NotoCBDT"])
    pdf.add_page()
    pdf.multi_cell(w=pdf.epw, text=text)
    assert_pdf_equal(pdf, HERE / "cbdt_noto_emoji_text.pdf", tmp_path)



import pytest
import sys

from pathlib import Path
from fpdf import FPDF
from test.conftest import assert_pdf_equal, check_signature, EPOCH

HERE = Path(__file__).resolve().parent
TRUSTED_CERT_PEMS = (HERE / "signing-certificate.crt",)


def test_sign_pkcs12(tmp_path):
    pdf = FPDF()
    pdf.set_creation_date(EPOCH)
    pdf.add_page()
    pdf.sign_pkcs12(HERE / "signing-certificate.p12", password=b"fpdf2")
    assert_pdf_equal(pdf, HERE / "sign_pkcs12.pdf", tmp_path)
    check_signature(pdf, TRUSTED_CERT_PEMS)


@pytest.mark.skipif(
    sys.platform in ("cygwin", "win32") and sys.version_info[:2] == (3, 14),
    reason="Skipped on Windows with Python 3.14 due to zlib compressed data differences",
)
def test_sign_pkcs12_with_link(tmp_path):
    "This test ensures that Signature & Link annotations can be combined"
    pdf = FPDF()
    pdf.set_creation_date(EPOCH)
    pdf.set_font("Helvetica", size=30)
    pdf.add_page()
    pdf.text(x=80, y=50, text="Page 1/2")
    pdf.add_page()
    pdf.set_xy(80, 50)
    pdf.sign_pkcs12(HERE / "signing-certificate.p12", password=b"fpdf2")
    pdf.cell(
        w=50,
        h=20,
        text="URL link",
        border=1,
        align="C",
        link="https://github.com/py-pdf/fpdf2",
    )
    assert_pdf_equal(pdf, HERE / "sign_pkcs12_with_link.pdf", tmp_path)
    check_signature(pdf, TRUSTED_CERT_PEMS)



from fpdf.graphics_state import GraphicsStateMixin

f = GraphicsStateMixin()
# Push initial state in stack: gs0
gs0 = f._push_local_stack()
# Step 1 - set some graphic styles: gs1
f.font_size_pt = 16
f.underline = True
gs1 = f._get_current_graphics_state()
# Step 2 - restore gs0
f._pop_local_stack()
print(f"{f.font_size_pt=} {f.underline=}")
# -> f.font_size_pt=0 f.underline=False



from pathlib import Path

from fpdf import FPDF
from test.conftest import LOREM_IPSUM, assert_pdf_equal, EMOJI_TEST_TEXT

HERE = Path(__file__).resolve().parent
FONTS_DIR = HERE.parent / "fonts"


def test_twemoji(tmp_path):
    # Twemoji - Mozilla build of the Twitter emojis on COLR format
    # Apache 2.0 license
    # https://github.com/mozilla/twemoji-colr
    pdf = FPDF()
    pdf.add_font("Twemoji", "", HERE / "Twemoji.Mozilla.ttf")
    pdf.add_page()
    test_text = "😂❤🤣👍😭🙏😘🥰😍😊"
    pdf.set_font("helvetica", "", 24)
    pdf.cell(text="Twemoi (COLRv0)", new_x="lmargin", new_y="next")
    pdf.cell(text="Top 10 emojis:", new_x="right", new_y="top")
    pdf.set_font("Twemoji", "", 24)
    pdf.cell(text=test_text, new_x="lmargin", new_y="next")
    assert_pdf_equal(pdf, HERE / "colrv0-twemoji.pdf", tmp_path)


def test_twemoji_shaping(tmp_path):
    pdf = FPDF()
    pdf.add_font("Twemoji", "", HERE / "Twemoji.Mozilla.ttf")
    pdf.add_page()
    combined_emojis = "🇫🇷 🇺🇸 🇨🇦 🧑 🧑🏽 🧑🏿"
    pdf.set_font("helvetica", "", 24)
    pdf.cell(text="Emojis without text shaping:", new_x="lmargin", new_y="next")
    pdf.set_font("Twemoji", "", 24)
    pdf.multi_cell(w=pdf.epw, text=combined_emojis, new_x="lmargin", new_y="next")
    pdf.ln()
    pdf.set_font("helvetica", "", 24)
    pdf.cell(text="Emojis with text shaping:", new_x="lmargin", new_y="next")
    pdf.set_font("Twemoji", "", 24)
    pdf.set_text_shaping(True)
    pdf.multi_cell(w=pdf.epw, text=combined_emojis, new_x="lmargin", new_y="next")
    assert_pdf_equal(pdf, HERE / "colrv0-twemoji_shaping.pdf", tmp_path)


def test_twemoji_text(tmp_path):
    text = EMOJI_TEST_TEXT
    pdf = FPDF()
    pdf.add_font("Roboto", "", FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("Twemoji", "", HERE / "Twemoji.Mozilla.ttf")
    pdf.set_font("Roboto", "", 24)
    pdf.set_fallback_fonts(["Twemoji"])
    pdf.add_page()
    pdf.multi_cell(w=pdf.epw, text=text)
    assert_pdf_equal(pdf, HERE / "colrv0-twemoji_text.pdf", tmp_path)


def test_noto_colrv1(tmp_path):
    # Twemoji - Mozilla build of the Twitter emojis on COLR format
    # Apache 2.0 license
    # https://github.com/mozilla/twemoji-colr
    pdf = FPDF()
    pdf.add_font("NotoColrv1", "", HERE / "colrv1-NotoColorEmoji.ttf")
    pdf.add_page()
    test_text = "😂❤🤣👍😭🙏😘🥰😍😊"
    pdf.set_font("helvetica", "", 24)
    pdf.cell(text="NotoColor (COLRv1)", new_x="lmargin", new_y="next")
    pdf.cell(text="Top 10 emojis:", new_x="right", new_y="top")
    pdf.set_font("NotoColrv1", "", 24)
    pdf.cell(text=test_text, new_x="lmargin", new_y="next")
    assert_pdf_equal(pdf, HERE / "colrv1-noto-color-emoji.pdf", tmp_path)


def test_noto_colrv1_shaping(tmp_path):
    pdf = FPDF()
    pdf.add_font("NotoColrv1", "", HERE / "colrv1-NotoColorEmoji.ttf")
    pdf.add_page()
    combined_emojis = "🇫🇷 🇺🇸 🇨🇦 🧑 🧑🏽 🧑🏿"
    pdf.set_font("helvetica", "", 24)
    pdf.cell(text="Emojis without text shaping:", new_x="lmargin", new_y="next")
    pdf.set_font("NotoColrv1", "", 24)
    pdf.multi_cell(w=pdf.epw, text=combined_emojis, new_x="lmargin", new_y="next")
    pdf.ln()
    pdf.set_font("helvetica", "", 24)
    pdf.cell(text="Emojis with text shaping:", new_x="lmargin", new_y="next")
    pdf.set_font("NotoColrv1", "", 24)
    pdf.set_text_shaping(True)
    pdf.multi_cell(w=pdf.epw, text=combined_emojis, new_x="lmargin", new_y="next")
    assert_pdf_equal(pdf, HERE / "colrv1-noto-color-emoji_shaping.pdf", tmp_path)


def test_noto_colrv1_text(tmp_path):
    text = EMOJI_TEST_TEXT
    pdf = FPDF()
    pdf.add_font("Roboto", "", FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("NotoColrv1", "", HERE / "colrv1-NotoColorEmoji.ttf")
    pdf.set_font("Roboto", "", 24)
    pdf.set_fallback_fonts(["NotoColrv1"])
    pdf.add_page()
    pdf.multi_cell(w=pdf.epw, text=text)
    assert_pdf_equal(pdf, HERE / "colrv1-noto-color-emoji_text.pdf", tmp_path)


def test_colrv1_bungee(tmp_path):
    # Bungee Color - OFL license
    # https://github.com/djrrb/Bungee

    pdf = FPDF()
    pdf.add_font("Bungee", "", HERE / "BungeeSpice-Regular-COLRv1.ttf")

    pdf.add_page()
    pdf.set_font("Bungee", size=16)
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.upper(), align="L")
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.upper(), align="R")
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.upper(), align="J")

    assert_pdf_equal(pdf, HERE / "colrv1_bungee.pdf", tmp_path)


def test_colrv1_nabla(tmp_path):
    pdf = FPDF()
    pdf.add_font("Nabla", "", HERE / "Nabla-Regular-COLRv1-VariableFont_EDPT,EHLT.ttf")

    pdf.add_page()
    pdf.set_font("Nabla", size=16)
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.upper(), align="L")
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.upper(), align="R")
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.upper(), align="J")

    assert_pdf_equal(pdf, HERE / "colrv1_nabla.pdf", tmp_path)
